{% include "header.html" %}
<ul class="object-tools">
    <li><a href="add/">{{ add_link }}</a></li>
</ul>

<div class="module">
    <table width="100%">
        <tr>
{% for field in field_list %}
            <th>{{ field.name }}</th>
{% endfor %}
            <th>Actions</th>
        </tr>
{% for object in object_list %}
        <tr id="{{ forloop.counter }}">
{% for value in object.value_list %}
            <td>{% if forloop.counter == "1" %}<a href="{{ object.pk }}/">{{ value }}</a>{% else %}{{ value }}{% endif %}</td>
{% endfor %}
            <td>
                <a href="{{ object.pk }}/">{{ edit_link }}</a>
                <a href="{{ object.pk }}/delete/">{{ delete_link }}</a>
            </td>
        </tr>
{% endfor %}
    </table>
</div>
{% include "footer.html" %}
